""" Generators for basis evaluations """

from dune.perftool.generation import (cached,
                                      domain,
                                      generator_factory,
                                      iname,
                                      instruction,
                                      preamble,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadrature_point,
                                             quadrature_iname,
                                             )
from dune.perftool.pdelab.geometry import (name_jacobian_inverse_transposed,
                                           )


@preamble('blubb')
def define_lfs_bound(lfs, bound):
    return 'auto {} = {}.size();'.format(bound, lfs)


@symbol
def name_lfs_bound(lfs):
    bound = '{}_size'.format(lfs)
    define_lfs_bound(lfs, bound)

    return bound


@generator_factory(cache_key_generator=lambda e, **kw: e, item_tags=('blubb',))
def name_lfs(element, prefix=None):
    # Omitting the prefix is only valid upon a second call, which will
    # result in a cache hit.
    assert prefix

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            # TODO in this case, we need to trigger the extraction mechanism
            # as preambles in our code.
            name_lfs(subelem, prefix + "_" + str(i))

    # Now trigger the creation of all those other symbols/preables necessary for this lfs

    # Now return the prefix!
    return prefix


def traverse_lfs_tree(arg):
    from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
    assert isinstance(arg, ModifiedArgumentDescriptor)

    # First we need to determine the basename as given in the signature of
    # this kernel method!
    basename = None
    from ufl.classes import Argument, Coefficient
    if isinstance(arg.argexpr, Argument):
        if arg.argexpr.count() == 0:
            basename = 'lfsv'
        if arg.argexpr.count() == 1:
            basename = 'lfsu'
        # TODO add restrictions here.

    if isinstance(arg.argexpr, Coefficient):
        # We should only ever call this for a trialfunction, which in our case
        # is the coefficient of reserved index 0.
        assert arg.argexpr.count() == 0

        basename = 'lfsu'

    assert basename

    # Now start recursively extracting local function spaces and fill the cache with
    # all those values. That way we can later get a correct local function space with
    # just the ufl finite element.
    name_lfs(arg.argexpr.element(), prefix=basename)


@iname
def _lfs_iname(element, argcount):
    name = name_lfs(element)
    bound = name_lfs_bound(name)

    if argcount != 0:
        name = 'lfsu'

    name = name + '_index'
    domain(name, bound)

    return name


def lfs_iname(element, argcount=0):
    return _lfs_iname(element, argcount)


@symbol
def name_localfunctionspace(expr):
    return 'lfsu'


@cached
def evaluate_basis(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n',))
    lfs = name_localfunctionspace(element)
    qp = name_quadrature_point()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{}.finiteElement().localBasis().evaluateFunction({}, {});'.format(lfs,
                                                                                        qp,
                                                                                        name,
                                                                                        ),
                assignees=name,
                )


@symbol
def name_basis(element):
    evaluate_basis(element, "phi")
    return "phi"


@cached
def evaluate_reference_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n', 2))
    lfs = name_localfunctionspace(element)
    qp = name_quadrature_point()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{}.finiteElement().localBasis().evaluateJacobian({}, {});'.format(lfs,
                                                                                        qp,
                                                                                        name,
                                                                                        ),
                assignees=name,
                )


@symbol
def name_reference_gradient(element):
    evaluate_reference_gradient(element, "js")
    return "js"


@cached
def evaluate_basis_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n', 2))
    jac = name_jacobian_inverse_transposed()
    index = lfs_iname(element)
    reference_gradients = name_reference_gradient(element)
    instruction(inames=(index,
                        quadrature_iname(),
                        ),
                code='{}.mv({}[{}][0], {});'.format(jac,
                                                    reference_gradients,
                                                    index,
                                                    name,
                                                    ),
                assignees=name,
                )


@symbol
def name_basis_gradient(element):
    evaluate_basis_gradient(element, "gradphi")
    return "gradphi"


@cached
def evaluate_trialfunction(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=())
    lfs = name_localfunctionspace(element)
    index = lfs_iname(element)
    basis = name_basis()
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{} += x({}, {})*{}[{}];'.format(name,
                                                      lfs,
                                                      index,
                                                      basis,
                                                      index
                                                      ),
                assignees=name,
                )


@cached
def evaluate_trialfunction_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=(2,))
    lfs = name_localfunctionspace(element)
    index = lfs_iname(element)
    basis = name_basis_gradient(element)
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{}.axpy(x({}, {}), {});'.format(name,
                                                      lfs,
                                                      index,
                                                      basis
                                                      ),
                assignees=name,
                )
